//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: C++ Client Service Example
//
// Description:
//
// Implementation
//
// Date:   Wed 05/17/2007
// Revision: $Revision: 1.4 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:13 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/ClientService/CHMclientService.cpp,v $
//---------------------------------------------------------------------------
#include "CHMclientServicePrecomp.h"
#pragma hdrstop

#include <CHM/CHMllpClient.h>
#include <SERV/SERVostream.h>
#include <SERV/SERVevent.h>
#include <SERV/SERVutils.h>
#include <SERV/SERVsinkStandard.h>
#include "CHMclientService.h"
#include "CHMconstants.h"
#include "CHMclient.h"

/// This application is a simple example of creating a Windows Service using the Chameleon 
/// 'Client' configuration.  One thing to keep in mind though is that the networking components
/// supplied with Chameleon require a Windows 'message loop'.  It is intended to be a helpful starting 
/// point in understanding how to use these objects, but must be fleshed out in the following 
/// areas to provide a full solution for an application HL7 interface:
/// 1) You will need to work out how your application will handle the logging of messages
///    and data, in particular how bad transactions should be logged for effective troubleshooting.
/// 2) The code will need to be modified to insert the parsed data into a database, whether
///    by using explicit database calls or by going through some type of middle tier layer that
///    your application may support.
/// 3) Definitely take the time to read the manual and go through the tutorials after looking
///    at this example and before attempting to deploy a production solution.  We do provide custom
///    on-site training and technical implementation services if you require them.
/// See the manual for more information about running Chameleon within services.
/// 
/// If you are looking for a much more complete 'out of the box' solution for your HL7 needs it 
/// would be a good idea to evaluate iNTERFACEWARE Iguana.  Iguana is powered by Chameleon, but
/// provides a complete run-time web-based infrastructure for your HL7 interfaces.
/// It provides advanced logging and monitoring capability which makes it a very attractive 
/// option for many customers.
/// 
/// It should be strongly emphasized that this example program is only for demonstration purposes.
/// A real HL7 interface should have some type of system of logging into log files.  Iguana, for
/// instance, has a very nice logging system that allows administrators to drill into and browse
/// processed messages and errors.  How you implement this logging is up to your requirements
/// for run-time management of your HL7 interface.

CHMboolean CHMclientService::onInit(CHMboolean RunningAsService)
{
   try
   {
      IsRunningAsService = RunningAsService;

      if (IsRunningAsService)
      {
         CHMstring LogFilePath;
         SERVgetExecutableFileDirectory(LogFilePath);
         LogFilePath += APP_LOG_FILE;

         LogFile.open(LogFilePath.c_str(), SERVbinaryFile::Rewrite);

         // Redirect output to the file.
         SERVcerr.setSink(LogFile);
         SERVcerr.setNewLine(APP_NEW_LINE);
         SERVcout.setSink(LogFile);
         SERVcout.setNewLine(APP_NEW_LINE);
      }

      return true;
   }
   catch (CHMerror& Error)
   {
      CHMstring Description("Error while initializing the service: ");
      Description += Error.Description();

      SERVsystemLog(APP_SERVICE_NAME, Description.c_str(), SERV_ERROR);
      return false;
   }
   catch (...)
   {
      SERVsystemLog(APP_SERVICE_NAME, "Unhandled exception while initializing server.", SERV_ERROR);
   }
}

void CHMclientService::onRun()
{
   try
   {
      CHMclient Client;
      pClient = &Client;

      SERVcout << "Running the service." << newline << flush;

      Client.startTimer(APP_TIMER_INTERVAL);

      // Need to call runMessageLoop() to start up the client.
      // Note that this is a blocking call.
      CHMllpClient::RunMessageLoop();

      pClient = NULL;

      if (IsRunningAsService)
      {
         // Redirect output back.
         SERVcerr.setSink(new SERVsinkStandardErr(), true);
         SERVcout.setSink(new SERVsinkStandardOut(), true);

         LogFile.close();
      }
   }
   catch (CHMerror& Error)
   {
      pClient = NULL;
      SERVcerr << "Error while running the service: " << Error.Description() << newline << flush;
   }
   catch (...)
   {
      pClient = NULL;
      SERVcerr << "Unhandled exception while running the service." << newline << flush;
   }
}

void CHMclientService::onStop()
{
   try
   {
      SERVcout << "Stopping the service." << newline << flush;

      if (pClient != NULL)
      {
         // Kill the timer.
         pClient->killTimer();

         // Stop the message loop.
         pClient->StopMessageLoop();
      }
   }
   catch (CHMerror& Error)
   {
      SERVcerr << "Error while stopping the service: " << Error.Description() << newline << flush;
   }
   catch (...)
   {
      SERVcerr << "Unhandled exception while stopping the service." << newline << flush;
   }
}
