//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVbinaryFile
//
// Description:
//
// Implementation
//
// Date:   Fri 01/16/2004 
// Revision: $Revision: 1.3 $
//
// Last Edit Date: $Date: 2007/06/20 16:21:20 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVbinaryFile.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#ifdef WIN32
#   include <io.h>
#   define SERVlseek   _lseek
#   define SERVlseek64 _lseeki64
#   define SERVopen    _open
#   define SERVfileno  _fileno
#   define SERVclose   _close
#   define SERVwrite   _write
#   define SERVread    _read
#   define SERVcommit  _commit
#   define SERVeof     _eof
#else
#   include <unistd.h>  
#   include <sys/types.h>
#   define O_BINARY 0
#   define SERVlseek   lseek
#   define SERVlseek64 lseek
#   define SERVopen    open
#   define SERVfileno  fileno
#   define SERVclose   close
#   define SERVwrite   write
#   define SERVread    read
#   define SERVcommit  fsync
#endif

#include <CHM/CHMminimumInclude.h>
#include <CHM/CHMerrorClass.h>
#include <CHM/CHMstring.h>
#include "SERVbinaryFile.h"
#include "SERVdebugLog.h"
#include "SERVutils.h"

#define SERVnpos SERVfileIndex(-1)

static void SERVdisableFileHandleInheritance(int FileDescriptor, const CHMstring& FileName)
{
#ifdef WIN32
   // now try to prohibit inheritance of this handle by a child process
   HANDLE WindowsFileHandle = (HANDLE)_get_osfhandle(FileDescriptor);
   if ((long)WindowsFileHandle == -1)
   {
      SERV_THROW_ERROR("Cannot get windows handle for file " + FileName, 0);
   }
   DWORD Flags;
   if (!GetHandleInformation(WindowsFileHandle, &Flags))
   {
      SERV_THROW_WIN32_ERROR("trying to get the handle information in SERVdisableFileHandleInheritance()");
   }
   Flags &= (~HANDLE_FLAG_INHERIT);
   if (!SetHandleInformation(WindowsFileHandle, HANDLE_FLAG_INHERIT, Flags))
   {
      SERV_THROW_WIN32_ERROR("trying to set the handle information in SERVdisableFileHandleInheritance()");
   }
#endif
}

class SERVbinaryFilePrivate
{
public:
   SERVbinaryFilePrivate() {}
   virtual ~SERVbinaryFilePrivate() {}
   virtual void open(const CHMstring& FileName,
                     SERVbinaryFile::SERVmode Mode)=0;
  
   virtual void close()=0;
   
   virtual CHMboolean isOpen() const=0;

   virtual SERVfileIndex position() const=0;
   virtual void setPosition(SERVfileIndex NewPosition)=0;

   virtual CHMuint32 read(void* pBuffer, CHMuint32 SizeOfBuffer)=0;
   virtual CHMuint32 write(const void* pBuffer, CHMuint32 SizeOfBuffer)=0;

   virtual SERVfileIndex size() const=0;

   virtual void flush()=0;

   virtual CHMboolean isEndOfSource() const=0;

#ifdef WIN32
   virtual HANDLE win32Handle() = 0;
#endif

   CHMstring FileName;
};

class SERVbinaryFilePrivateBuffered 
: public SERVbinaryFilePrivate
{
public:
   SERVbinaryFilePrivateBuffered(); 
    
   virtual ~SERVbinaryFilePrivateBuffered();

   virtual void open(const CHMstring& FileName,
                     SERVbinaryFile::SERVmode Mode);
  
   virtual void close();
   
   virtual CHMboolean isOpen() const;

   virtual SERVfileIndex position() const;
   virtual void setPosition(SERVfileIndex NewPosition);

   virtual CHMuint32 read(void* pBuffer, CHMuint32 SizeOfBuffer);
   virtual CHMuint32 write(const void* pBuffer, CHMuint32 SizeOfBuffer);

   virtual SERVfileIndex size() const;

   virtual void flush();

   virtual CHMboolean isEndOfSource() const;

   FILE* FileHandle;
   CHMstring Option;
   mutable SERVfileIndex FileSize;

   enum SERVlastOperation
   {
      FLUSH,
      READ,
      WRITE
   };
   
#ifdef WIN32
   virtual HANDLE win32Handle();
#endif

   SERVlastOperation LastOperation;
};

SERVbinaryFilePrivateBuffered::SERVbinaryFilePrivateBuffered()
: SERVbinaryFilePrivate(),
  FileHandle(NULL),
  FileSize(SERVnpos),
  LastOperation(FLUSH)
{

}

SERVbinaryFilePrivateBuffered::~SERVbinaryFilePrivateBuffered()
{
   try
   {
      close();
   }
   catch (const CHMerror& Error)
   {
      SERV_LOG(Error);
   }
}

CHMboolean SERVbinaryFilePrivateBuffered::isEndOfSource() const
{
   CHMPRECONDITION(FileHandle != NULL);
   return (0 != feof(FileHandle));
}

void SERVbinaryFilePrivateBuffered::open
(
   const CHMstring&        iFileName,
   SERVbinaryFile::SERVmode Mode 
)
{
   close();
   LastOperation = FLUSH;
   FileName = iFileName;
   FileSize = SERVnpos;

   switch(Mode)
   {
   case SERVbinaryFile::Read:
      Option = "rb"; 
      break;    // open for reading only
   case SERVbinaryFile::Write:   // open for reading and writing
      Option = "rb+";
      break;
   case SERVbinaryFile::Append:  // open and seek to the end
      Option = "ab+";
      break;
   case SERVbinaryFile::Rewrite: // open, truncate any existing file
      Option = "wb+";
      break;
   default:
      SERV_THROW_ERROR("Unknown File Parameter", 0);
      break;
   };
   FileHandle = ::fopen(FileName.c_str(), Option.c_str());
   if (NULL == FileHandle)
   {   
      // We have a file error
      SERV_THROW_STRERROR("Open file '" + FileName + "' failed");
   }
   SERVdisableFileHandleInheritance( SERVfileno(FileHandle), FileName );
}

void SERVbinaryFilePrivateBuffered::close()
{
   if (NULL != FileHandle)
   {
      flush();
      if (EOF == fclose(FileHandle))
      {
         FileHandle = NULL;
         SERV_THROW_STRERROR("Closing file '" + FileName + "' failed");
      }
   }
   FileHandle = NULL;
   FileSize = SERVnpos;
}
   
CHMboolean SERVbinaryFilePrivateBuffered::isOpen() const
{
   return FileHandle != NULL;
}

SERVfileIndex SERVbinaryFilePrivateBuffered::position() const
{
   CHMPRECONDITION(FileHandle != NULL);
   long Position = ftell(FileHandle);
   if (Position == SERVnpos)
   {
      SERV_THROW_STRERROR("ftell failed on '" + FileName + "'");
   }
   return SERVfileIndex(Position);
}

void SERVbinaryFilePrivateBuffered::setPosition(SERVfileIndex NewPosition)
{
   CHMPRECONDITION(FileHandle != NULL);
   
   if (0 != fseek(FileHandle, (long)NewPosition, SEEK_SET))
   {
      char NewPositionBuffer[33];
      CHMstring NewPositionString(_ui64toa(NewPosition, NewPositionBuffer, 10));
      SERV_THROW_STRERROR("fseek failed to move to " + NewPositionString + " on '" + FileName + "'");
   }
}

CHMuint32 SERVbinaryFilePrivateBuffered::read(void* pBuffer, CHMuint32 SizeOfBuffer)
{
   CHMPRECONDITION(FileHandle != NULL);
   if (LastOperation == WRITE)
   {
      flush();
   }
   LastOperation = READ;

   CHMuint32 AmountRead = fread(pBuffer, 1, SizeOfBuffer, FileHandle);
   if (AmountRead < SizeOfBuffer)
   {
      // We need to check to see if we have an error or if we simply have reached the
      // end of the file
      if (ferror(FileHandle))
      {
         SERV_THROW_STRERROR("fread failed on '" + FileName + "'");
      }
   }
   return AmountRead;
}

CHMuint32 SERVbinaryFilePrivateBuffered::write(const void* pBuffer, CHMuint32 SizeOfBuffer)
{
   CHMPRECONDITION(FileHandle != NULL);
   if (SizeOfBuffer == 0)
   {
      return SizeOfBuffer;
   }
   if (LastOperation == READ)
   {
      flush();
   }
   LastOperation = WRITE;

   FileSize = SERVnpos;
   // initialize the value of errno before calling fwrite
   errno = 0;
   size_t CountWritten = fwrite(pBuffer, 1, SizeOfBuffer, FileHandle );
   if (CountWritten == SizeOfBuffer)
   {
      return SizeOfBuffer; // we are good
   }
   // We may have got to the end of the file.  Our desired behavior
   // in this case is to move the file position to before the end of
   // the file and quietly recover
   if (errno == 0)   // fwrite has not changed the value of errno
   {
      setPosition(size());
      FileSize = SERVnpos;
      CountWritten = fwrite(pBuffer, 1, SizeOfBuffer, FileHandle);
      if (CountWritten == SizeOfBuffer)
      {
         return SizeOfBuffer;
      }
   }
   SERV_THROW_STRERROR("fwrite failed on '" + FileName + "'");
}

SERVfileIndex SERVbinaryFilePrivateBuffered::size() const
{
   CHMPRECONDITION(FileHandle != NULL);
   if (FileSize == SERVnpos)
   {
      // We haven't got a cached value of the file size
      SERVfileIndex OldPosition = position();

      if (0 != fseek(FileHandle, 0, SEEK_END))
      {
         SERV_THROW_STRERROR("fseek during size() call failed on '" + FileName + "'");
      }
      FileSize = position();
      ((SERVbinaryFilePrivateBuffered*)this)->setPosition(OldPosition);
   }
   return FileSize;
}

void SERVbinaryFilePrivateBuffered::flush()
{
   CHMPRECONDITION(FileHandle != NULL);
   if (Option == "rb" )
   {
       // this is read-only file so we cannot flush
       return;
   }
   if (fflush(FileHandle) != 0)
   {
      SERV_THROW_STRERROR("fflush failed on '" + FileName + "'");
   }
   LastOperation = FLUSH;
}

#ifdef WIN32
HANDLE SERVbinaryFilePrivateBuffered::win32Handle()
{
   if (!FileHandle)
   {
      return 0;
   }
   return (HANDLE)_get_osfhandle( SERVfileno(FileHandle) );
}
#endif

class SERVbinaryFilePrivateUnbuffered 
: public SERVbinaryFilePrivate
{
public:
   SERVbinaryFilePrivateUnbuffered();
   virtual ~SERVbinaryFilePrivateUnbuffered();
   virtual void open(const CHMstring& FileName,
                     SERVbinaryFile::SERVmode Mode);
  
   virtual void close();
   
   virtual CHMboolean isOpen() const;

   virtual SERVfileIndex position() const;
   virtual void setPosition(SERVfileIndex NewPosition);

   virtual CHMuint32 read(void* pBuffer, CHMuint32 SizeOfBuffer);
   virtual CHMuint32 write(const void* pBuffer, CHMuint32 SizeOfBuffer);

   virtual SERVfileIndex size() const;

   virtual void flush();

   virtual CHMboolean isEndOfSource() const;

#ifdef WIN32
   virtual HANDLE win32Handle();
#endif

   int FileHandle;
   int OptionFlag;
   mutable SERVfileIndex FileSize;
};

SERVbinaryFilePrivateUnbuffered::SERVbinaryFilePrivateUnbuffered()
: SERVbinaryFilePrivate(),
  FileSize(SERVnpos),
  OptionFlag(-1),
  FileHandle(-1){}

SERVbinaryFilePrivateUnbuffered::~SERVbinaryFilePrivateUnbuffered()
{
   try
   {
      close();
   }
   catch (const CHMerror& Error)
   {
      SERV_LOG(Error);
   }
}

void SERVbinaryFilePrivateUnbuffered::open
(
   const CHMstring& iFileName,
   SERVbinaryFile::SERVmode Mode
)
{
   close();
   FileName = iFileName;
   FileSize = SERVnpos;

   switch(Mode)
   {
   case SERVbinaryFile::Read:
      OptionFlag = O_BINARY | O_RDONLY; 
      break;    // open for reading only
   case SERVbinaryFile::Write:   // open for reading and writing
      OptionFlag = O_BINARY | O_RDWR ;
      break;
   case SERVbinaryFile::Append:  // open and seek to the end
      OptionFlag = O_BINARY | O_RDWR | O_CREAT | O_APPEND;
      break;
   case SERVbinaryFile::Rewrite: // open, truncate any existing file
      OptionFlag = O_BINARY | O_RDWR | O_CREAT | O_TRUNC;
      break;
   default:
      SERV_THROW_ERROR("Unknown File Parameter", 0);
      break;
   };
   if (SERVbinaryFile::Read == OptionFlag)
   {
      FileHandle = ::SERVopen(FileName.c_str(), OptionFlag);
   }
   else
   {
      FileHandle = ::SERVopen(FileName.c_str(), OptionFlag, S_IREAD | S_IWRITE);
   }
   if (-1 == FileHandle)
   {
      // we have a file error
      SERV_THROW_STRERROR("Open file '" + FileName + "' failed");
   }  
   SERVdisableFileHandleInheritance( FileHandle, FileName );
}

void SERVbinaryFilePrivateUnbuffered::close()
{
   if (-1 != FileHandle)
   {
      if (-1 == ::SERVclose(FileHandle))
      {
         FileHandle = -1;
         SERV_THROW_STRERROR("Closing file '" + FileName + "' failed");
      }
   }
   FileHandle = -1;
   FileSize = SERVnpos;
}

CHMboolean SERVbinaryFilePrivateUnbuffered::isEndOfSource() const
{
   CHMPRECONDITION(FileHandle != -1);

#ifdef WIN32
   int ReturnCode = SERVeof(FileHandle);
   if (ReturnCode == -1)
   {
      SERV_THROW_STRERROR("_eof failed");
   }
   return (ReturnCode == 1);
#else
   off_t CurrentPosition = lseek(FileHandle, 0, SEEK_CUR);
   off_t EndOfFilePosition = lseek(FileHandle, 0, SEEK_END);
   lseek(FileHandle, CurrentPosition, SEEK_SET);

   return (CurrentPosition == EndOfFilePosition);
#endif
}
   
CHMboolean SERVbinaryFilePrivateUnbuffered::isOpen() const
{
   return FileHandle != -1;
}

SERVfileIndex SERVbinaryFilePrivateUnbuffered::position() const
{
   CHMPRECONDITION(FileHandle != -1);
   // permit files of up to (2^64) - 1 bytes
   CHMuint64 Position = SERVlseek64(FileHandle, 0, SEEK_CUR);
   if (Position == -1)
   {
      SERV_THROW_STRERROR("lseek failed on '" + FileName + "'");
   }
   return SERVfileIndex(Position);
}

void SERVbinaryFilePrivateUnbuffered::setPosition(SERVfileIndex NewPosition)
{
   CHMPRECONDITION(FileHandle != -1);
   
   if (-1 == SERVlseek64(FileHandle, NewPosition, SEEK_SET))
   {
      char NewPositionBuffer[33];
      CHMstring NewPositionString(_ui64toa(NewPosition, NewPositionBuffer, 10));
      SERV_THROW_STRERROR("lseek failed to move to " + NewPositionString + " on '" + FileName + "'");
   }
}

CHMuint32 SERVbinaryFilePrivateUnbuffered::read(void* pBuffer, CHMuint32 SizeOfBuffer)
{
   CHMPRECONDITION(FileHandle != -1);
   CHMuint32 AmountRead = ::SERVread(FileHandle, pBuffer, SizeOfBuffer);
   
   if (AmountRead == -1 )
   {
      SERV_THROW_STRERROR("read failed on '" + FileName + "'");
   }
   return AmountRead;
}

CHMuint32 SERVbinaryFilePrivateUnbuffered::write(const void* pBuffer, CHMuint32 SizeOfBuffer)
{
   CHMPRECONDITION(FileHandle != -1);
   CHMuint32 AmountWritten = ::SERVwrite(FileHandle, pBuffer, SizeOfBuffer);

   if (AmountWritten == -1 || AmountWritten < SizeOfBuffer)
   {
      SERV_THROW_STRERROR("write failed on '" + FileName + "'");
   }
   FileSize = SERVnpos;

   return SizeOfBuffer;
}

SERVfileIndex SERVbinaryFilePrivateUnbuffered::size() const
{
   CHMPRECONDITION(FileHandle != -1);
   if (FileSize == SERVnpos)
   {
      // We haven't got a cached value of the file size
      SERVfileIndex OldPosition = position();

      FileSize = SERVlseek64(FileHandle, 0, SEEK_END);

      if (SERVnpos == FileSize)
      {
         SERV_THROW_STRERROR("lseek during size() call failed on '" + FileName + "'");
      }
      ((SERVbinaryFilePrivateUnbuffered*)this)->setPosition(OldPosition);
   }
   return FileSize;
}

void SERVbinaryFilePrivateUnbuffered::flush()
{
   if ((OptionFlag & O_RDONLY) == O_RDONLY )
   {
      // This is read only file so we cannot flush
      return;
   }
   if (-1 == ::SERVcommit(FileHandle))
   {
      // This often happens if the file has not be opened for writing to it.
      SERV_THROW_STRERROR("commit flush called failed on '" + FileName + "'");
   }
}

#ifdef WIN32
HANDLE SERVbinaryFilePrivateUnbuffered::win32Handle()
{
   if (FileHandle == -1)
   {
      return 0;
   }
   return (HANDLE)_get_osfhandle(FileHandle);
}
#endif

SERVbinaryFile::SERVbinaryFile(const CHMstring& FileName, SERVmode Mode, SERVtype Type)
: pMember(NULL)
{
   SERV_LOG("SERVbinaryFile Constructor.");
   if (Buffered == Type)
   {
      pMember = new SERVbinaryFilePrivateBuffered();      
   }
   else
   {
      pMember = new SERVbinaryFilePrivateUnbuffered();      
   }
   try
   {
      SERV_LOG("About to open " << FileName);
      open(FileName, Mode);
   }
   catch(...)
   {
      SERV_LOG("Exception while constructing file.");
      delete pMember;
      SERV_LOG("Cleaned up after exception - about to re-throw error for application");
      throw;
   }
}  

SERVbinaryFile::SERVbinaryFile(SERVtype Type)
 : pMember(NULL)
{
   if (Buffered == Type)
   {
      pMember = new SERVbinaryFilePrivateBuffered();      
   }
   else
   {
      pMember = new SERVbinaryFilePrivateUnbuffered();      
   }
}

SERVbinaryFile::~SERVbinaryFile()
{
   delete pMember;
}

void SERVbinaryFile::open
(
    const CHMstring& FileName,
    SERVmode Mode  
) 
{
   pMember->open(FileName, Mode);
}
  
void SERVbinaryFile::close()
{
   pMember->close();
}
   
CHMboolean SERVbinaryFile::isOpen() const
{
   return pMember->isOpen();
}

SERVindex SERVbinaryFile::position() const
{
   CHMPRECONDITION(pMember->position() <= CHM_UINT_32_MAX);
   return (SERVindex)pMember->position();
}

void SERVbinaryFile::setPosition(SERVindex NewPosition)
{
   pMember->setPosition(NewPosition);
}

CHMuint64 SERVbinaryFile::position64() const
{
   return pMember->position();
}

void SERVbinaryFile::setPosition64(CHMuint64 NewPosition)
{
   pMember->setPosition(NewPosition);
}

CHMuint32 SERVbinaryFile::read(void* pBuffer, CHMuint32 SizeOfBuffer)
{
   return pMember->read(pBuffer, SizeOfBuffer);
}

CHMuint32 SERVbinaryFile::write(const void* pBuffer, CHMuint32 SizeOfBuffer)
{
   return pMember->write(pBuffer, SizeOfBuffer);
}

SERVfileIndex SERVbinaryFile::size() const
{
   return pMember->size();
}

const CHMstring& SERVbinaryFile::name() const
{
   return pMember->FileName;
}

void SERVbinaryFile::flush()
{
   pMember->flush();
}

void SERVbinaryFile::setName(const CHMstring& NewName)
{
   pMember->FileName = NewName;
}

#ifdef WIN32
HANDLE SERVbinaryFile::win32Handle()
{
   return pMember->win32Handle();
}
#endif
