//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVevent
//
// Description:
//
// Implementation
//
// Date:   Friday, May 18th, 2007 @ 04:15:54 PM
// Revision: $Revision: 1.4 $
//
// Last Edit Date: $Date: 2007/06/20 16:21:20 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVevent.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include "SERVmessages.h" // Special generated file from SERVmessages.mc - see custom build step.
#include "SERVserver.h"
#include "SERVevent.h"
#include "SERVutils.h"

#include <CHM/CHMstring.h>
#include <CHM/CHMerrorClass.h>

void SERVaddEventSource(const CHMstring& SourceName, const CHMstring& ServiceFilePath)
{
   CHMstring SubKey("SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\");
   SubKey += SourceName;

   HKEY CurrentKey = NULL;

   // Check if user can modify.
   if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, 
                      SubKey.c_str(), 
                      0, 
                      "", 
                      REG_OPTION_NON_VOLATILE, 
                      KEY_ALL_ACCESS, NULL, 
                      &CurrentKey, 
                      NULL) != ERROR_SUCCESS )
   {
      // Key probably cannot be created. try to open read only.
      SERVcheckResult(RegOpenKeyEx(HKEY_LOCAL_MACHINE, SubKey.c_str(), 0, KEY_READ, &CurrentKey));
   }

   SERVcheckResult(RegSetValueExA(CurrentKey, "EventMessageFile", 0, REG_SZ, (LPBYTE)ServiceFilePath.c_str(), ServiceFilePath.size()));

   CHMint32 Value = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE;
   SERVcheckResult(RegSetValueEx(CurrentKey, "TypesSupported", 0, REG_DWORD, (LPBYTE)&Value, sizeof(Value)));
}

void SERVaddEventSource(const SERVserver& Server)
{
   CHMstring ExeNamePath;
   SERVfullExecutableFileName(ExeNamePath);
   SERVaddEventSource(Server.name(), ExeNamePath);
}

void SERVremoveEventSource(const CHMstring& SourceName)
{
   CHMstring SubKey("SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\");

   HKEY CurrentKey = NULL;

   // Check if user can modify.
   if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, 
                      SubKey.c_str(), 
                      0, 
                      "", 
                      REG_OPTION_NON_VOLATILE, 
                      KEY_ALL_ACCESS, NULL, 
                      &CurrentKey, 
                      NULL) != ERROR_SUCCESS )
   {
      // Key probably cannot be created. Try to open read only.
      SERVcheckResult(RegOpenKeyEx(HKEY_LOCAL_MACHINE, SubKey.c_str(), 0, KEY_READ, &CurrentKey));
   }

   LONG Result = RegDeleteKey(CurrentKey, SourceName.c_str());

   if ((Result != ERROR_SUCCESS) && (Result != ERROR_FILE_NOT_FOUND))
   {
       SERV_THROW_ERROR(SERVformatWin32Error(Result).c_str(), 0);
   }
}

void SERVremoveEventSource(const SERVserver& Server)
{
   SERVremoveEventSource(Server.name());
}

// Converts the severity enum to a WIN32 event type.
WORD SERVsystemLogSeverityToEventType(SERVsystemLogSeverity Severity)
{   
   switch (Severity)
   {
   case SERV_EMERGENCY:
   case SERV_ALERT:
   case SERV_CRITICAL:
   case SERV_ERROR:
      return EVENTLOG_ERROR_TYPE;
   case SERV_WARNING:
      return EVENTLOG_WARNING_TYPE;
   case SERV_NOTICE:
   case SERV_INFO:
   case SERV_DEBUG:
   default:
      return EVENTLOG_INFORMATION_TYPE;
   }
}

void SERVsystemLog(const CHMstring& SourceName, const CHMstring& LogLine, SERVsystemLogSeverity Severity)
{
   HANDLE EventLogHandle = ::RegisterEventSource(NULL, SourceName.c_str());
   if (NULL == EventLogHandle)
   {
      SERV_THROW_WIN32_ERROR("trying to register the event source in CHMsystemLog()");
   }

   const char* MessageString = LogLine.c_str();
   if (!::ReportEvent(EventLogHandle,                  // event log handle 
                      SERVsystemLogSeverityToEventType(Severity), // event type 
                      0,                               // category zero 
                      CHM_EVENT_MESSAGE,               // event identifier 
                      NULL,                            // no user security identifier 
                      1,                               // one substitution string 
                      0,                               // no data 
                      (LPCTSTR*) &MessageString,       // pointer to string array 
                      NULL))                           // pointer to data 
   {
      SERV_THROW_WIN32_ERROR("trying to report the event in CHMsystemLog()");
   }

   if (0 == ::DeregisterEventSource(EventLogHandle))
   {
      SERV_THROW_WIN32_ERROR("trying to deregister the event source in CHMsystemLog()");
   }
}

void SERVsystemLog(const SERVserver& Server, const CHMstring& LogLine, SERVsystemLogSeverity Severity)
{
   SERVsystemLog(Server.name(), LogLine, Severity);
}
