//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVhash
//
// Description:
//
// Implementation
//
// Date:   March 2004 
// Revision: $Revision: 1.3 $
//
// Last Edit Date: $Date: 2007/06/20 16:21:20 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVhash.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include <CHM/CHMstring.h>
#include "SERVdebugLog.h"
#include "SERVhash.h"

#define SERV_LOOKUP_HASH_HASH_VALUE_BITS 32

// CHMstring specialization
template <>
SERVindex SERVhash<CHMstring>::defaultHash(const CHMstring& Key)
{
   if (!Key.size())
   {
      return 0;
   }
   else
   {
      SERV_LOG("SERVhash<CHMstring>::defaultHash...");
      //TODO - replace all one letter variables with meaningful names
      //TODO - add some comments for block of code, eliminate magic values
      register CHMuint32 HashIncrement;
      CHMuint32 Hash = 0;
      CHMuint32 ByteIncrement = sizeof(CHMuint32);
      CHMuint32 ByteIndex;
      CHMuint32 c;
      CHMuint32 DividingByte = Key.size() & 0xFFFC; // the largest multiple of 4 <= Key.size()
      // On the Solaris and HP-UX platforms it is necessary to make a copy of Key before hashing,
      // otherwise a Bus Error and Core Dump can result - the reason for this is unknown
#ifdef CHM_SOLARIS
      CHMstring Key2(Key.c_str());
      const char* pStr = Key2.c_str();
#elif CHM_HPUX
      CHMstring Key2(Key.c_str());
      const char* pStr = Key2.c_str();       
#else
      const char* pStr = Key.c_str();
#endif
      SERV_LOG("Hash = " << Hash << "; DividingByte = " << DividingByte << "; Key = " << Key);
      // process 4 byte chunks of Key up to the DividingByte, hashing each into a 32 bit integer
      for (ByteIndex = 0; ByteIndex < DividingByte; ByteIndex += ByteIncrement)
      {
         HashIncrement = *(CHMuint32*) (pStr + ByteIndex);
         Hash += HashIncrement;
         c = (int(HashIncrement >> 24) + (int(HashIncrement >> 16) & 0x00FF) + (int(HashIncrement >> 8) & 0x00FF) + (int(HashIncrement) & 0x00FF)) % (SERV_LOOKUP_HASH_HASH_VALUE_BITS - 1) + 1;
         Hash = (Hash << (SERV_LOOKUP_HASH_HASH_VALUE_BITS - c)) | (Hash >> c);
      }
      // process bytes remaining after DividingByte
      for (ByteIndex = DividingByte; ByteIndex < Key.size(); ByteIndex++)
      {
         Hash += pStr[ByteIndex];
         c = (pStr[ByteIndex] + ByteIndex) % (SERV_LOOKUP_HASH_HASH_VALUE_BITS - 1) + 1;
         Hash = (Hash << (SERV_LOOKUP_HASH_HASH_VALUE_BITS - c)) | (Hash >> c);
      }
      return Hash;   
   }
}

//A Basic CRC hash we can use for other types
static const CHMuint32 SERV_LOOKUP_HASH_BYTE_WIDTH = 8; // Number of bits in a byte
static const CHMuint32 SERV_LOOKUP_HASH_WORD_WIDTH = sizeof (CHMuint32) * SERV_LOOKUP_HASH_BYTE_WIDTH;
static const CHMuint32 SERV_LOOKUP_HASH_CRC_HASH_SHIFT = 5;

inline CHMuint32 SERVhashCRCHash(void* Key, CHMuint32 Size)
{
   if(!Size)
   {
      return 0;
   }
   CHMuint32 HashValue = 0;
   CHMuint32 Index = 0;
   while (Index < (Size-1)) //note that 0 is always returned if Size = 1, so do not use for chars
   {
      CHMuint32 leftShiftedValue = HashValue << SERV_LOOKUP_HASH_CRC_HASH_SHIFT;
      CHMuint32 rightShiftedValue = HashValue >> (SERV_LOOKUP_HASH_WORD_WIDTH - SERV_LOOKUP_HASH_CRC_HASH_SHIFT);
   
      // Put the shifted values together, and then XOR them with the
      // next key character
      HashValue = (leftShiftedValue | rightShiftedValue) ^ (((CHMuint8*)Key)[Index+1]);
      Index++;
   }
   return HashValue;
}

static const CHMuint32 SizeofCHMuint64 = sizeof(CHMuint64);
static const CHMuint32 SizeofCHMuint32 = sizeof(CHMuint32);
static const CHMuint32 SizeofCHMuint16 = sizeof(CHMuint16);
static const CHMuint32 SizeofCHMuint8 = sizeof(CHMuint8);

//Integer specializations
template <>
SERVindex SERVhash<CHMuint64>::defaultHash(const CHMuint64& Key)
{
   return SERVhashCRCHash((void*)&Key,SizeofCHMuint64);
}

template <>
SERVindex SERVhash<CHMint64>::defaultHash(const CHMint64& Key)
{
   return SERVhashCRCHash((void*)&Key,SizeofCHMuint64);
}

template <>
SERVindex SERVhash<CHMuint32>::defaultHash(const CHMuint32& Key)
{
   return SERVhashCRCHash((void*)&Key,SizeofCHMuint32);
}

template <>
SERVindex SERVhash<CHMint32>::defaultHash(const CHMint32& Key)
{
   return SERVhashCRCHash((void*)&Key,SizeofCHMuint32);
}

template <>
SERVindex SERVhash<CHMuint16>::defaultHash(const CHMuint16& Key)
{
   return SERVhashCRCHash((void*)&Key,SizeofCHMuint16);
}

template <>
SERVindex SERVhash<CHMint16>::defaultHash(const CHMint16& Key)
{
   return SERVhashCRCHash((void*)&Key,SizeofCHMuint16);
}

//TODO, explain this better, get rid of magic value 3
//chars use the integer multiply algorithm
inline CHMuint32 SERVhashIntegerMultiply(CHMuint32 Key)
{
   return Key * (Key + 3);
}

template <>
SERVindex SERVhash<CHMuint8>::defaultHash(const CHMuint8& Key)
{
   return SERVhashIntegerMultiply(Key);
}

template <>
SERVindex SERVhash<CHMint8>::defaultHash(const CHMint8& Key)
{
   return SERVhashIntegerMultiply(Key);
}
