#ifndef __SERV_HASH_H__
#define __SERV_HASH_H__
//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVhash
//
// Description:
//
// The default hash functions. It is a template specialized to allow easy specification
// for our simple types.
//
// This is an example of using it:
//
// SERVhashTable<CHMstring, CHMstring> StudentTable(SERVhash<CHMstring>::defaultHash);
//
// Of course if you need more specialized hash function you can always write them
// and use SERVhashTable independently of SERVhash.
//
// Date:   March 2004 
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVhash.h,v $
//---------------------------------------------------------------------------
#include <CHM/CHMminimumInclude.h>

//supported default hash implementations
//
// CHMuint32/int32
// CHMuint64/int64
// CHMuint16/int16
// CHMuint8/int8  <- this usually is unsigned char/char
// CHMstring

typedef size_t SERVindex;

template <class T>
struct SERVhash
{
   static SERVindex defaultHash(const T& Key);
   
   //default compare defers the the types equal operator
   static CHMboolean equalKey(const T& Key1, const T& Key2)
   {
      return (Key1 == Key2) != 0;
   }
};

template <>
SERVindex SERVhash<CHMuint64>::defaultHash(const CHMuint64& Key);
template <>
SERVindex SERVhash<CHMint64>::defaultHash(const CHMint64& Key);
template <>
SERVindex SERVhash<CHMuint32>::defaultHash(const CHMuint32& Key);
template <>
SERVindex SERVhash<CHMint32>::defaultHash(const CHMint32& Key);
template <>
SERVindex SERVhash<CHMuint16>::defaultHash(const CHMuint16& Key);
template <>
SERVindex SERVhash<CHMint16>::defaultHash(const CHMint16& Key);
template <>
SERVindex SERVhash<CHMuint8>::defaultHash(const CHMuint8& Key);
template <>
SERVindex SERVhash<CHMint8>::defaultHash(const CHMint8& Key);

#ifndef CHM_64

template <typename T>
struct SERVhashPointer
{
   static SERVindex defaultHash(const T& Key)
   {
      return SERVhash<CHMuint32>::defaultHash((CHMuint32)Key);
   }

   static CHMboolean equalKey(const T& Key1, const T& Key2)
   {
      return (Key1 == Key2) != 0;
   }
};

#else

template <typename T>
struct SERVhashPointer
{
   static SERVindex defaultHash(const T& Key)
   {
      return SERVhash<CHMuint64>::defaultHash((CHMuint64)Key);
   }

   static CHMboolean equalKey(const T& Key1, const T& Key2)
   {
      return (Key1 == Key2) != 0;
   }
};

#endif

#endif // end of defensive include
