//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVostream
//
// Description:
//
// Implementation
//
// Date:   Fri 01/16/2000 
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVostream.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include <CHM/CHMerrorClass.h>
#include <CHM/CHMstring.h>
#include "SERVostream.h"
#include "SERVsink.h"
#include "SERVsinkStandard.h"

SERVostream SERVcerr(new SERVsinkStandardErr(), true);
SERVostream SERVcout(new SERVsinkStandardOut(), true);
SERVostream SERVclog(new SERVsinkStandardOut(), true);

class NumberFormats
{
public:
   NumberFormats()
   {
      NumberFormatsDecimal[eInt32] = "%i\0";
      NumberFormatsDecimal[eUint32] = "%u\0";
      NumberFormatsHex[eInt32] = "%X\0";
      NumberFormatsHex[eUint32] = "%X\0";

      NumberFormatsDecimal[eInt16] = "%hi\0";
      NumberFormatsDecimal[eUint16] = "%hu\0";
      NumberFormatsHex[eInt16] = "%hX\0";
      NumberFormatsHex[eUint16] = "%hX\0";

      NumberFormatsDecimal[eFloat32] = "%f\0";
      NumberFormatsDecimal[eDouble64] = "%f\0";
      NumberFormatsHex[eFloat32] = NumberFormatsDecimal[eFloat32];
      NumberFormatsHex[eDouble64] = NumberFormatsDecimal[eDouble64];

      #ifdef WIN32
         NumberFormatsDecimal[eInt64] = "%I64i\0";
         NumberFormatsDecimal[eUint64] = "%I64u\0";
         NumberFormatsHex[eInt64] = "%I64X\0";
         NumberFormatsHex[eUint64] = "%I64X\0";
      #else
         NumberFormatsDecimal[eInt64] = "%lli\0";
         NumberFormatsDecimal[eUint64] = "%llu\0";
         NumberFormatsHex[eInt64] = "%llX\0";
         NumberFormatsHex[eUint64] = "%llX\0";
      #endif
   }

   virtual ~NumberFormats(){}

   static const NumberFormats& instance()
   {
      static const NumberFormats Instance;
      return Instance;
   }

   enum DecimalNumberFormatType
   {
      eInt32 = 0,
      eUint32 = 1,
      eInt16 = 2,
      eUint16 = 3,
      eInt64 = 4,
      eUint64 = 5,
      eFloat32 = 6,
      eDouble64 = 7
   };
   enum {CountOfNumberFormat = 8};

   const char* numberFormatHex(DecimalNumberFormatType Format) const
   {
      CHMPRECONDITION(Format < CountOfNumberFormat);
      return NumberFormatsHex[Format];      
   }

   const char* numberFormatDecimal(DecimalNumberFormatType Format) const
   {
      CHMPRECONDITION(Format < CountOfNumberFormat);
      return NumberFormatsDecimal[Format];      
   }

private:
   const char* NumberFormatsDecimal[CountOfNumberFormat];
   const char* NumberFormatsHex[CountOfNumberFormat];
};

class SERVostreamPrivate
{
public:
   SERVostreamPrivate(SERVsink* ipSink, CHMboolean Owner) 
    : NewLine("\n"),
      NewLineAndIndent("\n"),
      IndentContent("   "),
      CountOfIndent(0),
      pSink(ipSink),
      NumberBase(SERVostream::Decimal),
      IsSinkOwner(Owner)
   {
      Buffer[BufferSize-1] = '\0'; //must null terminate
   }
   ~SERVostreamPrivate(){}

   const CHMchar* numberFormat(NumberFormats::DecimalNumberFormatType Format);
   SERVostream::ENumberBase NumberBase;
   SERVsink* pSink;

   CHMstring NewLine;
   CHMstring IndentContent;
   CHMstring NewLineAndIndent;
   CHMuint32 CountOfIndent;

   CHMboolean IsSinkOwner;

   void makeIndentString()
   {
      NewLineAndIndent = NewLine;
      for (CHMuint32 IndentIndex = 0; IndentIndex < CountOfIndent;IndentIndex++)
      {
         NewLineAndIndent += IndentContent;
      }
   }

   enum {BufferSize = 100};//maximum

   CHMchar Buffer[BufferSize];
};

const char* SERVostreamPrivate::numberFormat(NumberFormats::DecimalNumberFormatType Format)
{
   if( NumberBase == SERVostream::Hexadecimal )
   {
      return NumberFormats::instance().numberFormatHex(Format);
   }
   else
   {
      return NumberFormats::instance().numberFormatDecimal(Format);
   }
}

SERVostream::SERVostream(SERVsink* pSink, CHMboolean Owner)
{
   pMember = new SERVostreamPrivate(pSink, Owner);
}

SERVostream::SERVostream(SERVsink& Sink)
{
   pMember = new SERVostreamPrivate(&Sink, false); // not the owner
}

SERVostream::~SERVostream()
{
   if (pMember->IsSinkOwner)
   {
      delete pMember->pSink;
   }
   delete pMember;
}

void SERVostream::newline()
{
   pMember->pSink->write((CHMuint8*)pMember->NewLineAndIndent.c_str(), pMember->NewLineAndIndent.size());      
}

void SERVostream::setNewLine(const char* NewLine)
{
   pMember->NewLine = NewLine;
   pMember->makeIndentString();
}

const char* SERVostream::getNewLine() const
{
   return pMember->NewLine.c_str();
}

void SERVostream::setIndentContent(const CHMstring& IndentContent)
{
   pMember->IndentContent = IndentContent;
   pMember->makeIndentString();
}

const CHMstring& SERVostream::indentContent() const
{
   return pMember->IndentContent;
}

void SERVostream::indent()
{
   pMember->CountOfIndent++;
   pMember->makeIndentString();
}

void SERVostream::unindent()
{
   if (pMember->CountOfIndent > 0)
   {
      pMember->CountOfIndent--;
   }
   pMember->makeIndentString();
}

void SERVostream::resetIndent()
{
   setCountOfIndent(0);
}

void SERVostream::setCountOfIndent(CHMuint32 NewValue)
{
   pMember->CountOfIndent = NewValue;
   pMember->makeIndentString();
}

CHMuint32 SERVostream::countOfIndent() const
{
   return pMember->CountOfIndent;
}

void SERVostream::flush()
{
   pMember->pSink->flush();
}

void SERVostream::setfill(CHMuint32 n)
{

}

CHMboolean SERVostream::isSinkOwner()
{
   return pMember->IsSinkOwner;
}

SERVsink* SERVostream::sink()
{
   return pMember->pSink;
}  

SERVsink* SERVostream::setSink(SERVsink* pSink, CHMboolean IsOwner)
{
   CHMPRECONDITION(pSink != NULL);
   if (pMember->IsSinkOwner)
   {
      delete pMember->pSink;
   }
   pMember->pSink = pSink;
   pMember->IsSinkOwner = IsOwner;
   return pSink;
}

SERVostream::ENumberBase SERVostream::numberBase()
{
   return pMember->NumberBase;
}

void SERVostream::setNumberBase( ENumberBase Base )
{
   pMember->NumberBase = Base;
}

SERVostream& SERVostream::operator<<(SERVostream&(*Stream)(SERVostream&))
{
   Stream(*this);
   return *this;
}

SERVostream& SERVostream::operator<<(SERVostream&(*Stream)(SERVostream&, int n))
{
   return *this;
}

SERVostream& SERVostream::operator<<(unsigned char Character)
{
   pMember->pSink->write(&Character, 1);
   return *this;
}

SERVostream& SERVostream::operator<<(CHMchar Character)
{
   pMember->pSink->write((CHMuint8*)&Character,1);
   return *this;
}

SERVostream& SERVostream::operator<<(const CHMchar* pString)
{
   if (pString)
   {
      pMember->pSink->write((CHMuint8*)pString, strlen(pString));
   }
   return *this;
}

SERVostream& SERVostream::operator <<(const CHMstring& String)
{
   pMember->pSink->write(String.c_str(), String.size());
   return *this;
}

SERVostream& SERVostream::operator<<(CHMint32 Number)
{
   int CountOfChar = sprintf(pMember->Buffer,pMember->numberFormat(NumberFormats::eInt32), Number);
   CHMPRECONDITION( CountOfChar < SERVostreamPrivate::BufferSize );
   pMember->pSink->write((CHMuint8*)pMember->Buffer, CountOfChar);
   return *this;
}

SERVostream& SERVostream::operator<<(CHMuint32 Number)
{
   int CountOfChar = sprintf(pMember->Buffer,pMember->numberFormat(NumberFormats::eUint32), Number);
   CHMPRECONDITION( CountOfChar < SERVostreamPrivate::BufferSize );
   pMember->pSink->write((CHMuint8*)pMember->Buffer, CountOfChar);
   return *this;
}

SERVostream& SERVostream::operator<<(CHMint16 Number)
{
   int CountOfChar = sprintf(pMember->Buffer,pMember->numberFormat(NumberFormats::eInt16), Number);
   CHMPRECONDITION( CountOfChar < SERVostreamPrivate::BufferSize );
   pMember->pSink->write((CHMuint8*)pMember->Buffer, CountOfChar);
   return *this;
}

SERVostream& SERVostream::operator<<(CHMuint16 Number)
{
   int CountOfChar = sprintf(pMember->Buffer,pMember->numberFormat(NumberFormats::eUint16), Number);
   CHMPRECONDITION( CountOfChar < SERVostreamPrivate::BufferSize );
   pMember->pSink->write((CHMuint8*)pMember->Buffer, CountOfChar);
   return *this;
}

SERVostream& SERVostream::operator<<(CHMint64 Number)
{
   int CountOfChar = sprintf(pMember->Buffer,pMember->numberFormat(NumberFormats::eInt64), Number);
   CHMPRECONDITION( CountOfChar < SERVostreamPrivate::BufferSize );
   pMember->pSink->write((CHMuint8*)pMember->Buffer, CountOfChar);
   return *this;
}

SERVostream& SERVostream::operator<<(CHMuint64 Number)
{
   int CountOfChar = sprintf(pMember->Buffer,pMember->numberFormat(NumberFormats::eUint64), Number);
   CHMPRECONDITION( CountOfChar < SERVostreamPrivate::BufferSize );
   pMember->pSink->write((CHMuint8*)pMember->Buffer, CountOfChar);
   return *this;
}

SERVostream& SERVostream::operator<<(CHMfloat32 Number)
{
   int CountOfChar = sprintf(pMember->Buffer,pMember->numberFormat(NumberFormats::eFloat32), Number);
   CHMPRECONDITION( CountOfChar < SERVostreamPrivate::BufferSize );
   pMember->pSink->write((CHMuint8*)pMember->Buffer, CountOfChar);
   return *this;
}

SERVostream& SERVostream::operator<<(CHMfloat64 Number)
{
   CHMuint8 GlobalFloat64Buffer[ 0x200 ] = {0};
   int CountOfChar = sprintf( (CHMint8*)GlobalFloat64Buffer,pMember->numberFormat(NumberFormats::eDouble64), Number);
   CHMPRECONDITION( CountOfChar < sizeof( GlobalFloat64Buffer ) );
   pMember->pSink->write(GlobalFloat64Buffer, CountOfChar);
   return *this;
}

#ifndef CHM_NO_BOOLEAN_TYPE
SERVostream& SERVostream::operator <<(CHMboolean Boolean)
{
   if(Boolean == true)
   {
      return operator<<("True");
   }
   return operator<<("False");
}
#endif

SERVostream& hex(SERVostream& _outs)
{
   _outs.setNumberBase(SERVostream::Hexadecimal);
   return _outs;
}

SERVostream& decimal(SERVostream& _outs)
{
   _outs.setNumberBase(SERVostream::Decimal);
   return _outs;
}


void __set_indent(SERVostream& OutStream, const CHMstring& IndentContent)
{
   OutStream.setIndentContent(IndentContent);
}

void __setfill(SERVostream& OutStream, const CHMchar& FillChar)
{
   OutStream.setfill(FillChar);
}

SERVostream& indent(SERVostream& OutStream)
{
   OutStream.indent();
   return OutStream;
}

SERVostream& unindent(SERVostream& OutStream)
{
   OutStream.unindent();
   return OutStream;
}

SERVostream& reset_indent(SERVostream& OutStream)
{
   OutStream.resetIndent();
   return OutStream;
}

SERVostreamManip* set_indent(const CHMstring& TagName)
{
   return new SERVostreamManip1<CHMstring>(__set_indent, TagName);
}

SERVostreamManip* setfill(CHMchar FillChar)
{
   return new SERVostreamManip1<CHMchar>(__setfill, FillChar);
}
