#ifndef __SERV_SERVER_H__
#define __SERV_SERVER_H__
//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVserver
//
// Description:
//
// This class is an abstraction of a windows server.  An application written 
// using this class will be compiled as a service under Windows.  Virtual 
// methods are used to be overridden by applications using this class.
//
// Date:   Wed 05/17/2007
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVserver.h,v $
//---------------------------------------------------------------------------

#include <CHM/CHMminimumInclude.h>

class CHMstring;
class SERVserverPrivate;

class SERVserver
{
public:
   SERVserver(const CHMstring& ServiceName);
   virtual ~SERVserver();

   const CHMstring& name() const;

   // This command will run the server as a Service under windows.  It will 
   // return false if the process is started from the command line - 
   // i.e. indicating that the process is not being started as a server.
   CHMboolean runServer();

   //  This command will run the server as a normal command line process.
   void runServerAsNormalProcess();

   // Method that is called before running the server i.e. provides facility to initialize a server.
   // No exceptions must be thrown from this overridden method.
   // Return true if initialization was successful.
   virtual CHMboolean onInit(CHMboolean RunningAsService) = 0;

   // Method that is meant to perform functionality of the server.
   // It should not exit until onStop is called.
   // No exceptions must be thrown from this overridden method.
   virtual void onRun() = 0;
  
   // This method must be over-ridden and is invoked when the server stops.
   // Exceptions must not be thrown by application code inheriting this method.
   virtual void onStop() = 0;

   void notifyStartProblem(CHMuint32 ErrorCode);

private:
   SERVserverPrivate* pMember;
   SERVserver(const SERVserver& Orig); // Copy
   SERVserver& operator=(const SERVserver& Orig); // Assignment
};

#endif // end of defensive include
