//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVservice
//
// Description:
//
// Implementation
//
// Date:   Friday, May 18th, 2007 @ 04:34:04 PM
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVservice.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include <CHM/CHMstring.h>
#include <CHM/CHMerrorClass.h>
#include "SERVdebugLog.h"
#include "SERVservice.h"
#include "SERVserviceManager.h"
#include "SERVutils.h"

class SERVservicePrivate
{
public:
   SERVservicePrivate(SERVserviceManager& MyManager, CHMstring iSourceName)
    : pQueryInfoBuffer(NULL),
      SizeOfBuffer(0),
      ServiceValidState(false),
      ServiceHandle(NULL) {}
   ~SERVservicePrivate() {}

   // The following two methods are used to abstract the need to know
   // the size of the buffer. Different systems might have different buffer
   // sizes. The buffer size is in the QueryServiceConfig function.
   void createQueryInfoBuffer();
   void populateQueryInfoBuffer();

   DWORD SizeOfBuffer;
   char* pQueryInfoBuffer;

   QUERY_SERVICE_CONFIG* queryServiceConfig()
   {
      return ((QUERY_SERVICE_CONFIG*)pQueryInfoBuffer);
   }

   CHMboolean ServiceValidState;
   SC_HANDLE ServiceHandle;
};

void SERVservicePrivate::createQueryInfoBuffer()
{
   // A trick to get size of buffer
   // Note that where size should be, a 0 is input
   // This will cause the query to return the size needed
   // in sizeOfBuffer parameter which can the be used to populate the buffer
   ::QueryServiceConfig(ServiceHandle,
                        NULL, 
                        0,
                        &SizeOfBuffer);
   
   pQueryInfoBuffer = new char[SizeOfBuffer];

   if (NULL == ::memset( pQueryInfoBuffer, '0', SizeOfBuffer ))
   {
      SERV_THROW_WIN32_ERROR("trying to set buffer (memset) to character '0' in SERVservicePrivate::createQueryInfoBuffer()");
   }
}

void SERVservicePrivate::populateQueryInfoBuffer()
{
   CHMPRECONDITION(pQueryInfoBuffer != NULL);
   DWORD CountOfByteNeeded=0;

   if (0 == ::QueryServiceConfig(ServiceHandle,
                                 (QUERY_SERVICE_CONFIG*)pQueryInfoBuffer, 
                                 SizeOfBuffer,
                                 &CountOfByteNeeded))
   {
      SERV_THROW_WIN32_ERROR("trying to QUERY service in SERVservicePrivate::populateQueryInfoBuffer()");
   }
}

SERVservice::SERVservice
(
   SERVserviceManager& Manager,
   const CHMstring&   ServiceName,
   CHMboolean         Create,
   CHMboolean         StartAutomatically
)
: pMember(NULL)
{
   pMember = new SERVservicePrivate(Manager, ServiceName);

   if (Create)
   {      
      DWORD StartType = SERVICE_DEMAND_START;
      if (StartAutomatically)
      {
         StartType = SERVICE_AUTO_START;
      }

      pMember->ServiceHandle = ::CreateService(Manager.handle(),// handle to service control manager 
                                               ServiceName.c_str(), // pointer to name of service to start
                                               ServiceName.c_str(), // pointer to display name
                                               SERVICE_ALL_ACCESS,   // access rights
                                               SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS,  // type of service
                                               StartType,        // when to start service - start condition
                                               SERVICE_ERROR_NORMAL, // severity if service fails to start
                                               ServiceName.c_str(), // pointer to name of binary file
                                               NULL, // pointer to name of load ordering group
                                               NULL, // pointer to variable to get tag identifier
                                               NULL, // pointer to array of dependency names
                                               NULL,  // pointer to account name of service
                                               NULL); // pointer to password for service account

      if (NULL == pMember->ServiceHandle)
      {
          SERV_THROW_WIN32_ERROR("trying to CREATE service in SERVservice::SERVservice()");
      }
   }
   else
   {
      pMember->ServiceHandle = ::OpenService(Manager.handle(),// handle to service control manager 
                                             ServiceName.c_str(),  // pointer to name of service to start
                                             SERVICE_ALL_ACCESS); // type of access to service 

      if (NULL == pMember->ServiceHandle)
      {
          SERV_THROW_WIN32_ERROR("trying to OPEN service in SERVservice::SERVservice()");
      }
   }

   pMember->createQueryInfoBuffer();
   pMember->ServiceValidState = true;
}

SERVservice::~SERVservice()
{
   try
   {
      if (pMember->ServiceHandle != NULL)
      {
         if (0 == ::CloseServiceHandle(pMember->ServiceHandle))
         {
            delete pMember;
            SERV_THROW_WIN32_ERROR("trying to CLOSE service in SERVservice::~SERVservice()");
         }
      }

      delete pMember->pQueryInfoBuffer;  // free query info buffer (if NULL then nothing is freed) 
      delete pMember;
   }
   catch (const CHMerror& Error) 
   {
      SERV_LOG(Error);
   }
}

// This is the uninstall method of a service
void SERVservice::deleteService()
{
   CHMPRECONDITION(pMember->ServiceHandle != NULL);
   CHMPRECONDITION(pMember->ServiceValidState == true);

   if (0 == ::DeleteService(pMember->ServiceHandle))
   {
      SERV_THROW_WIN32_ERROR("trying to uninstall service in SERVservice::deleteService()");
   }
}

// The internal name of the service.
const CHMstring SERVservice::name() const 
{
   CHMPRECONDITION(pMember->ServiceValidState == true);
   pMember->populateQueryInfoBuffer();
   return pMember->queryServiceConfig()->lpServiceStartName; 
}

// The displayed name of the service.
const CHMstring SERVservice::displayName() const 
{
   CHMPRECONDITION(pMember->ServiceValidState == true);
   pMember->populateQueryInfoBuffer();
   return pMember->queryServiceConfig()->lpDisplayName; 
}

void SERVservice::setExecutableFile(const CHMstring& FileName) 
{
   CHMPRECONDITION(pMember->ServiceValidState == true);
   CHMPRECONDITION(!FileName.is_null());

   CHMstring FilePath ;
   CHMboolean HasEmbeddedSpace = false;

   for (CHMuint32 CharIndex = 0; CharIndex < FileName.size(); CharIndex++)
   { 
      // Check for white space, in which case quotes will have to be added.
      if (FileName[CharIndex] == ' '|| 
          FileName[CharIndex] == '\t'||
          FileName[CharIndex] == '\r'|| 
          FileName[CharIndex] == '\n'||
          FileName[CharIndex] == '\0'|| 
          FileName[CharIndex] == '\x0B')
      {
         HasEmbeddedSpace = true;
      }
   }

   // Quotes will appear in executable file name
   // Thats fine since that is how windows deals
   // with white spaces in filenames.
   if (HasEmbeddedSpace)
   {
      FilePath = "\"";
      FilePath += FileName;
      FilePath += "\"";
   }
   else 
   {
      FilePath = FileName;
   }

   if (0 == ::ChangeServiceConfig(pMember->ServiceHandle,
                                  SERVICE_NO_CHANGE,
                                  SERVICE_NO_CHANGE,
                                  SERVICE_NO_CHANGE,
                                  FilePath.c_str(),
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL))
   {
      SERV_THROW_WIN32_ERROR("while changing the executable file for a service to " + FilePath);
   }
}

const CHMstring SERVservice::executableFile() const
{
   pMember->populateQueryInfoBuffer();
   return pMember->queryServiceConfig()->lpBinaryPathName;
}

void SERVservice::start()
{ 
   CHMPRECONDITION(pMember->ServiceValidState == true);
   
   if (0 == ::StartService(pMember->ServiceHandle,
                           0,
                           NULL))
   {
      SERV_THROW_WIN32_ERROR("trying to START service in SERVservice::start()");
   } 
}

void SERVservice::stop()
{
   CHMPRECONDITION(pMember->ServiceValidState == true);

   SERVICE_STATUS Status;

   if (0 == ::ControlService(pMember->ServiceHandle,
                             SERVICE_CONTROL_STOP,
                             &Status))
   {
      SERV_THROW_WIN32_ERROR("trying to STOP service in SERVservice::stop()");
   } 
}

CHMboolean SERVservice::isRunning() const
{
   CHMPRECONDITION(pMember->ServiceValidState == true);

   SERVICE_STATUS Status;

   if (0 == ::QueryServiceStatus(pMember->ServiceHandle, // handle of service
                                 &Status)) // address of service status structure
   {
      SERV_THROW_WIN32_ERROR("trying to QUERY service status in SERVservice::isRunning()");
   }

   if (SERVICE_RUNNING == Status.dwCurrentState)
   {
      return true;
   }
   else
   {
      return false;
   }
}

void SERVservice::setDisplayName(const CHMstring& DisplayName)
{
   CHMPRECONDITION(!DisplayName.is_null());
   CHMPRECONDITION(pMember->ServiceValidState == true);

   if (0 == ::ChangeServiceConfig(pMember->ServiceHandle,
                                  SERVICE_NO_CHANGE,
                                  SERVICE_NO_CHANGE,
                                  SERVICE_NO_CHANGE,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  NULL,
                                  DisplayName.c_str()))
   {
      SERV_THROW_WIN32_ERROR("trying to change the service display name in SERVservice::setServiceName()");
   }
}

typedef BOOL (__stdcall * CHMchangeServiceConfig)(SC_HANDLE Handle, DWORD InfoLevel, LPVOID pInfo);

void SERVservice::setDescription(const CHMstring& ServiceDescription)
{
   CHMPRECONDITION(!ServiceDescription.is_null());
   CHMPRECONDITION(pMember->ServiceHandle != NULL);
   CHMPRECONDITION(pMember->ServiceValidState == true);

   HINSTANCE DllHandle = ::LoadLibrary("ADVAPI32.DLL");
   if (DllHandle == NULL)
   {
      SERV_THROW_WIN32_ERROR("trying to load the library ADVAPI32.DLL in SERVservice::setDescription()");
   }

   // ChangeServiceConfig2A in ADVAPI32.DLL exists on WinXP, Win2000 Pro, Win Server 2003 and Win2000 Server
   CHMchangeServiceConfig pChangeServiceFunction = (CHMchangeServiceConfig) ::GetProcAddress(DllHandle, "ChangeServiceConfig2A");

   // If we have an older operating system that does not support this functionality, 
   // we gracefully return and do not do anything.
   if (NULL != pChangeServiceFunction)
   {
      CHMint32 Result = -1;
      SERVICE_DESCRIPTION ServiceDescriptionStruct;
      ServiceDescriptionStruct.lpDescription = (char*) ServiceDescription.c_str();
   
      Result = (*pChangeServiceFunction)(pMember->ServiceHandle, SERVICE_CONFIG_DESCRIPTION, &ServiceDescriptionStruct);

      if (0 == Result)
      {
         SERV_THROW_WIN32_ERROR("trying to set the service description");
      }
   }

   if (0 == ::FreeLibrary(DllHandle))
   {
      SERV_THROW_WIN32_ERROR("trying to free the library ADVAPI32.DLL in SERVservice::setDescription()");
   }
}
