#ifndef __SERV_SERVICE_H__
#define __SERV_SERVICE_H__
//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVservice
//
// Description:
//
// This class encapsulates the access and control to a windows service. 
// It is not the actual service program - which is encapsulated in CHMserver.
//
// NOTE:
// This class represents a Service object - we do not construct 
// these classes directly - you should obtain references to them 
// from the SERVserviceManager class.
//
// Date:   Wed 05/17/2007
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVservice.h,v $
//---------------------------------------------------------------------------

class CHMstring;
class SERVservicePrivate;
class SERVserviceManager;

class SERVservice
{
public:
   // Note: The constructor has been declared as private in order to restrict creation 
   // of an instance to CHMserviceManager - which is a friend
   ~SERVservice();

   // Returns a pointer to the name (key) of the service.
   const CHMstring name() const;

   // Returns the display name of the service
   const CHMstring displayName() const;

   // Sets the display name of a service. Will throw a system error if change to service configuration 
   // is unsuccessful.
   void setDisplayName(const CHMstring& DisplayName);

   // Sets the description of the service. Is only applicable when used with Windows 2000 or later
   void setDescription(const CHMstring& ServiceDescription);

   // Sets path to executable of a service. Will throw a system error if change to service configuration 
   // is unsuccessful.  
   void setExecutableFile(const CHMstring& FileName);

   // Returns pointer to path of executable.
   const CHMstring executableFile() const;

   // Starts the service. Will throw an exception if service is not started successfully.
   void start();

   // Stops the service. Will throw an exception if service is not started successfully.
   void stop();

   // Returns bool to signify if service is started. Will throw an error if status query is unsuccessful.
   CHMboolean isRunning() const;

   friend class SERVserviceManager;

private:
   // This can only be invoked by the service handle.
   SERVservice(SERVserviceManager& Manager, 
              const CHMstring& ServiceName, 
              CHMboolean Create,
              CHMboolean StartAutomatically);

   // Implementation only - called by the SERVserviceManager
   void deleteService();

   SERVservicePrivate* pMember;
   SERVservice(const SERVservice& Orig);
   SERVservice& operator=(const SERVservice& Orig);
};

#endif // end of defensive include
