//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVserviceManager
//
// Description:
//
// Implementation
//
// Date:   Friday, May 18th, 2007 @ 04:16:01 PM
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVserviceManager.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include <CHM/CHMstring.h>
#include <CHM/CHMerrorClass.h>
#include "SERVhashTable.h"
#include "SERVserviceManager.h"
#include "SERVservice.h"
#include "SERVutils.h"

class SERVserviceManagerPrivate
{
public:
   SERVserviceManagerPrivate()
   : ManagerHandle(NULL) {}
   ~SERVserviceManagerPrivate();

   SC_HANDLE ManagerHandle;

   SERVhashTable<CHMstring, SERVservice*> ServiceTable;
};

SERVserviceManagerPrivate::~SERVserviceManagerPrivate()
{
   for (SERVhashTablePlace ServiceTablePlace = ServiceTable.first(); ServiceTablePlace != NULL; ServiceTablePlace = ServiceTable.next(ServiceTablePlace))
   {
      delete ServiceTable[ServiceTablePlace];
   }
}

SERVserviceManager::SERVserviceManager()
: pMember(NULL)
{
   pMember = new SERVserviceManagerPrivate();

   pMember->ManagerHandle = ::OpenSCManager(NULL, // using local machine so we leave this parameter null
                                            NULL, // use the default service database
                                            SC_MANAGER_ALL_ACCESS);

   if (NULL == pMember->ManagerHandle)
   {
      SERV_THROW_WIN32_ERROR("trying to open Service Control Manager in SERVserviceManager::SERVserviceManager()");
   }
}

SERVserviceManager::~SERVserviceManager()
{
   if (pMember->ManagerHandle != NULL)
   {
      // No point checking the return value here.
      ::CloseServiceHandle(pMember->ManagerHandle);
   }
   delete pMember;
}

CHMboolean SERVserviceManager::doesServiceExist(const CHMstring& ServiceName) const
{
   CHMPRECONDITION(pMember->ManagerHandle != NULL);

   SC_HANDLE ServiceHandle = ::OpenService(pMember->ManagerHandle,
                                           ServiceName.c_str(),
                                           SERVICE_QUERY_CONFIG);

   if (ServiceHandle) 
   {
      if (0 == ::CloseServiceHandle(ServiceHandle))
      {
         SERV_THROW_WIN32_ERROR("trying to close service handle in SERVserviceManager::doesServiceExist()");
      }

      return true;
   }
   else
   {
      return false;
   }
}

// If the service exists already then this function will throw an exception.
SERVservice& SERVserviceManager::createService(const CHMstring& ServiceName, CHMboolean StartAutomatically)
{
   if (pMember->ServiceTable.find(ServiceName))
   {
      // Service has been created already?
      SERV_THROW_ERROR(ServiceName + " has been created already.", 0);
   }

   pMember->ServiceTable[ServiceName] = new SERVservice(*this, ServiceName, true, StartAutomatically);
   return *pMember->ServiceTable[ServiceName];
}

void SERVserviceManager::removeService(const CHMstring& ServiceName)
{
   SERVservice Service(*this, ServiceName, false, false);
   Service.deleteService();
}

// If the service does not exist then this function will throw an exception.
SERVservice& SERVserviceManager::service(const CHMstring& ServiceName)
{
   if (!pMember->ServiceTable.find(ServiceName))
   {
      pMember->ServiceTable[ServiceName] = new SERVservice(*this, ServiceName, false, false);
   }

   return *pMember->ServiceTable[ServiceName];
}

void* SERVserviceManager::handle()
{
   return (void*)pMember->ManagerHandle;
}
