//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVtimer
//
// Description:
//
// Implementation
//
// Date:   Friday, May 18th, 2007 @ 04:16:06 PM
// Revision: $Revision: 1.2 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVtimer.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include "SERVhashTable.h"
#include "SERVtimer.h"
#include "SERVutils.h"

class SERVtimerStatic
{
public:
   typedef SERVhashTable<CHMuint32, SERVtimer*> TimerTable;

   SERVtimerStatic() {}
   ~SERVtimerStatic() {}

   TimerTable Timers;
};

class SERVtimerPrivate
{
public:
   SERVtimerPrivate(SERVtimer* NewParent);
   ~SERVtimerPrivate();

   void createTimer(CHMuint32 TimeoutInMilliseconds);
   void killTimer();

   static VOID CALLBACK TimerProcedure
   (
      HWND WindowHandle,
      UINT Message,
      UINT Event,
      DWORD Time
   );

   static SERVtimerStatic::TimerTable& timers();

private:
   SERVtimer* Parent;
   CHMuint32 TimerId;

   static SERVtimerStatic* pStaticInstance;
   static CHMuint32 CountOfTimers;
};

SERVtimerStatic* SERVtimerPrivate::pStaticInstance = NULL;
CHMuint32 SERVtimerPrivate::CountOfTimers = 0;

SERVtimerPrivate::SERVtimerPrivate(SERVtimer* NewParent)
: Parent(NewParent),
  TimerId(0)
{
   ++CountOfTimers;

   if (!pStaticInstance)
   {
      pStaticInstance = new SERVtimerStatic;
   }
}

SERVtimerPrivate::~SERVtimerPrivate()
{
   killTimer();

   --CountOfTimers;

   if (0 == CountOfTimers)
   {
      delete pStaticInstance;
      pStaticInstance = NULL;
   }
}

void SERVtimerPrivate::createTimer(CHMuint32 TimeoutInMilliseconds)
{
   TimerId = SetTimer(NULL, 0, TimeoutInMilliseconds, TimerProcedure);

   if (0 == TimerId)
   {
      SERV_THROW_WIN32_ERROR("setting the timer");
   }

   pStaticInstance->Timers[TimerId] = Parent;
}

void SERVtimerPrivate::killTimer()
{
   if (0 != TimerId)
   {
      // Destroy the timer.
      if (0 == KillTimer(NULL, TimerId))
      {
         SERV_THROW_WIN32_ERROR("killing the timer");
      }

      pStaticInstance->Timers.removeKey(TimerId);
      TimerId = 0;
   }
}

VOID CALLBACK SERVtimerPrivate::TimerProcedure
(
   HWND WindowHandle,
   UINT Message,
   UINT Event,
   DWORD Time
)
{
   // Find the correct timer.
   SERVhashTablePlace TimerPlace = SERVtimerPrivate::timers().find(Event);

   if (NULL != TimerPlace)
   {
      try
      {
         SERVtimerPrivate::timers()[TimerPlace]->onTimeout();
      }
      catch(...)
      {
         SERVexitCondition("C++ Exception thrown during onTimeout", __FILE__, __LINE__, "");
      }
   }
}

SERVtimerStatic::TimerTable& SERVtimerPrivate::timers()
{
   return pStaticInstance->Timers;
}

SERVtimer::SERVtimer()
{
   pMember = new SERVtimerPrivate(this);
}

SERVtimer::~SERVtimer()
{
   delete pMember;
}

void SERVtimer::startTimer(CHMuint32 TimeoutInMilliseconds)
{
   pMember->createTimer(TimeoutInMilliseconds);
}

void SERVtimer::killTimer()
{
   pMember->killTimer();
}
