//---------------------------------------------------------------------------
// Copyright (C) 1997-2007 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module: SERVutils
//
// Description:
//
// Implementation
//
// Date:   Friday, May 18th, 2007 @ 04:16:24 PM
// Revision: $Revision: 1.4 $
//
// Last Edit Date: $Date: 2007/06/15 20:58:14 $
// Source: $Source: /home/cvs/cvsroot/stubcode_templates/c++/SERV/SERVutils.cpp,v $
//---------------------------------------------------------------------------
#include "SERVprecomp.h"
#pragma hdrstop

#include <time.h>
#include <CHM/CHMerrorClass.h>
#include "SERVostream.h"
#include "SERVutils.h"
#include "SERVserviceManager.h"
#include "SERVservice.h"

void SERVfullExecutableFileName(CHMstring& FileName)
{
   FileName.assign(_MAX_FNAME, '\0');

   HMODULE FileHandle = ::GetModuleHandle(NULL);
   if (NULL == FileHandle)
   {
      SERV_THROW_WIN32_ERROR("trying to get the module handle");
   }

   DWORD CountOfChar = ::GetModuleFileName(FileHandle, FileName.get_buffer(), FileName.size());
   if (0 == CountOfChar)
   {
      SERV_THROW_WIN32_ERROR("trying to get the module file name");
   }

   FileName = FileName.substr(0, CountOfChar);
}

void SERVgetExecutableFileDirectory(CHMstring& Directory)
{
   SERVfullExecutableFileName(Directory);

   CHMuint32 CountOfChar = Directory.size();
   while ((CountOfChar > 0) && Directory[CountOfChar] != '\\')
   {
      --CountOfChar;
   }

   Directory = Directory.substr(0, CountOfChar + 1);
}

void SERVgetDateTimeAsString(CHMstring& DateTimeString)
{
   time_t CurrentTime;
   time(&CurrentTime);
   struct tm *Today = localtime(&CurrentTime);

   DateTimeString.assign(15, '\0');
   DateTimeString = DateTimeString.substr(0, strftime(DateTimeString.get_buffer(), 15, "%Y%m%d%H%M%S", Today));
}

CHMboolean SERVisServiceInstalled(const CHMstring& ServiceName)
{
   SERVserviceManager ServiceManager;
   return ServiceManager.doesServiceExist(ServiceName);
}

void SERVinstallService
(
   const CHMstring& ServiceName,
   const CHMstring& DisplayName,
   const CHMstring& Description,
   CHMboolean StartAutomatically
)
{
   SERVserviceManager ServiceManager;
   CHMstring ExecutableFileName;
   SERVfullExecutableFileName(ExecutableFileName);

   SERVcout << "Installing service..." << newline;

   if (ServiceManager.doesServiceExist(ServiceName)) 
   {
      SERVservice& Service = ServiceManager.service(ServiceName);
      CHMboolean IsRunning = Service.isRunning();

      if (IsRunning)
      {
         SERVcout << "Service is running, stopping it..." << newline;
         Service.stop();
         SERVcout << "Stopped.  You will need to restart the server." << newline;
      }

      Service.setExecutableFile(ExecutableFileName);
      Service.setDisplayName(DisplayName);
      Service.setDescription(Description);

      SERVcout << "Service re-installed." << newline << flush;
   }
   else
   {   
      SERVservice& Service = ServiceManager.createService(ServiceName, StartAutomatically); 
      Service.setExecutableFile(ExecutableFileName); 
      Service.setDisplayName(DisplayName);
      Service.setDescription(Description);

      SERVcout << "Service installed." << newline << flush;
   }
}

void SERVuninstallService(const CHMstring& ServiceName)
{
   CHMPRECONDITION(!ServiceName.is_null());

   SERVcout << "Uninstalling the service..." << newline;

   SERVserviceManager ServiceManager;

   if (ServiceManager.doesServiceExist(ServiceName))
   {
      if (ServiceManager.service(ServiceName).isRunning())
      {
         SERVcout << "Service is running.  Stopping service...";
         ServiceManager.service(ServiceName).stop();
         SERVcout << "Service stopped." << newline;
      }

      ServiceManager.removeService(ServiceName);
      SERVcout << "Service uninstalled." << newline << flush;
   }
   else
   {
      SERVcout << "Service was not installed so no change." << newline << flush;
   }
}

CHMstring SERVformatWin32Error(long ErrorCode)
{
   char* MessageBuffer;
   DWORD dwFormatFlags = FORMAT_MESSAGE_ALLOCATE_BUFFER |
                         FORMAT_MESSAGE_IGNORE_INSERTS | 
                         FORMAT_MESSAGE_FROM_SYSTEM;

   ::FormatMessage(dwFormatFlags,
      NULL, // module to get message from (NULL == system)
      ErrorCode,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // default language
      (char*) &MessageBuffer,
      0,
      NULL);

   CHMstring Result = MessageBuffer;
   LocalFree(MessageBuffer);
   return Result;
}

void SERVthrowSystemError
(
   const CHMstring& Description,
   CHMuint32 LineNumber,
   const char* FileName
)
{
   CHMstring ErrorMessage = "Error " + Description + ": " + SERVformatWin32Error(::GetLastError());
   throw CHMerror(ErrorMessage.c_str(), LineNumber, FileName, (CHMuint32)::GetLastError());
}

void SERVcheckResult(LONG Result)
{
   if (Result != ERROR_SUCCESS)
   {
       SERV_THROW_ERROR(SERVformatWin32Error(Result).c_str(), 0);
   }
}

static void SERVconsoleExitCondition
(
   const char* Condition,
   const char* FileName,
   CHMuint32 LineNo,
   const char* pDescription
)
{
   SERVcout << "FATAL EXIT CONDITION:" << newline
            << Condition << " not true in " << FileName << " at " << LineNo << newline;

   if (pDescription)
   {
      SERVcout << pDescription << newline;
   }

   SERVcout << "Press Enter to continue." << newline << flush;
   char c = getchar();

   exit(EXIT_FAILURE);
}

static void SERVwindowExitCondition
(
   const char* Condition,
   const char* FileName,
   CHMuint32 LineNo,
   const char* pDescription
)
{
   char LineNoBuffer[33];
   itoa(LineNo, LineNoBuffer, 10);

   CHMstring Message(Condition);
   Message += " not true in ";
   Message += FileName;
   Message += " at ";
   Message += LineNoBuffer;
   Message += "\r\n";

   if (pDescription)
   {
      Message += pDescription;
      Message += "\r\n";
   }

   ::MessageBox(NULL,  Message.c_str(), "FATAL EXIT CONDITION", MB_OK);

   exit(EXIT_FAILURE);
}

typedef HWND (*SERVgetHandleFunction)();

static CHMboolean SERVhasConsole()
{
   HINSTANCE DllHandle = ::LoadLibrary("kernel32.dll");
   if (DllHandle == NULL)
   {
       SERV_THROW_WIN32_ERROR("trying to load the library kernel32.dll in CHMhasConsole()");
   }

   SERVgetHandleFunction pFunction = (SERVgetHandleFunction) ::GetProcAddress(DllHandle, "GetConsoleWindow");

   if (pFunction != NULL)
   {
      HWND ConsoleHandle = (*pFunction)();
      return ConsoleHandle != 0;
   }
   else
   {
      // We have an older operating system than windows 2000 so
      // this call does not work.  Chances are we are not going
      // to have this code working in production on older windows OS.
      return false;
   }
}

void SERVexitCondition
(
   const char* Condition,
   const char* FileName,
   CHMuint32 LineNo,
   const char* pDescription
)
{
   if (SERVhasConsole())
   {
      // The program has a console.
      SERVconsoleExitCondition(Condition, FileName, LineNo, pDescription);
   }
   else
   {
      SERVwindowExitCondition(Condition, FileName, LineNo, pDescription);
   }
}
